//  Copyright 2016 root.ext@gmail.com
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import 'dart:typed_data';
import 'dart:mirrors';
import 'package:test/test.dart';

import '../lib/dna.dart';
import '../lib/src/native.dart';
import '../lib/src/coder.dart';
import '../lib/src/meta.dart';

void main() {
  group('tests', () {
    setUp(() {});

    tearDown(() {});

    test('call: params char', () {
      var lib = new TestLib();
      expect(lib.paramChar(0, 0), 0);
      expect(lib.paramChar(1, 2), 3);
      expect(lib.paramChar(-1, -2), -3);
      expect(lib.paramChar(127, 1), -128); //0x7F + 1 = 0x80
      expect(lib.paramChar(-2, 1), -1);    //0xFE + 1 = 0xFF
    });

    test('call: params unsigned char', () {
      var lib = new TestLib();
      expect(lib.paramUchar(0, 0), 0);
      expect(lib.paramUchar(1, 2), 3);
      expect(lib.paramUchar(127, 1), 128); //0x7F + 1 = 0x80
      expect(lib.paramUchar(254, 1), 255); //0xFE + 1 = 0xFF
    });
    
    test('call: params integer', () {
      var lib = new TestLib();
      expect(lib.paramInt(0, 0), 0);
      expect(lib.paramInt(1, 2), 3);
      expect(lib.paramInt(-1, -2), -3);
      expect(lib.paramInt(2147483647, 1), -2147483648); //0x7FFFFFFF + 1 = 0x80000000
      expect(lib.paramInt(-2, 1), -1);                  //0xFFFFFFFE + 1 = 0xFFFFFFFF
    });

    test('call: params unsigned integer', () {
      var lib = new TestLib();
      expect(lib.paramUint(0, 0), 0);
      expect(lib.paramUint(1, 2), 3);
      expect(lib.paramUint(2147483647, 1), 2147483648); //0x7FFFFFFF + 1 = 0x80000000
      expect(lib.paramUint(4294967294, 1), 4294967295); //0xFFFFFFFE + 1 = 0xFFFFFFFF
    });

    test('call: params float', () {
      var lib = new TestLib();
      expect(lib.paramFloat(0.0, 0.0) + 0.0, lessThan(0.000001));
      expect(lib.paramFloat(1.0, 2.0) - 3.0, lessThan(0.000001));
      expect(lib.paramFloat(-1.0, -2.0) + 3.0, lessThan(0.000001));
    });

    test('call: params double', () {
      var lib = new TestLib();
      expect(lib.paramDouble(0.0, 0.0) + 0.0, lessThan(0.000001));
      expect(lib.paramDouble(1.0, 2.0) - 3.0, lessThan(0.000001));
      expect(lib.paramDouble(-1.0, -2.0) + 3.0, lessThan(0.000001));
    });

    test('call: params bool', () {
      var lib = new TestLib();
      expect(lib.paramBool(false, false), false);
      expect(lib.paramBool(true, false), true);
      expect(lib.paramBool(false, true), true);
      expect(lib.paramBool(true, true), true);      
    });

    test('call: params string', () {
      var lib = new TestLib();
      
      expect(lib.paramString("abc"), 3);      
    });

    test('call: params string out', () {
      var lib = new TestLib();
      var param = new Ref("abc");

      expect(lib.paramStringOut(param), 3);
      expect(param.value, "zbc");
    });

    test('call: param struct', () {
      var lib = new TestLib();
      var res = lib.paramStruct(new IntsSt(1, 2, 3));
      expect(res, 6);
    });

    test('call: param struct pointer', () {
      var lib = new TestLib();
      var param = new IntsSt(1,2,3);
      var meta = getStructMeta(reflectType(IntsSt));
      var pointer = getPointer(encodeStruct(param, null, meta, [], []));
      var res = lib.paramStructPointer(pointer);
      expect(res, 6);
    });

    test('call: param struct out', () {
      var lib = new TestLib();
      var param = new IntsSt(1,2,3);
      var res = lib.paramStructOut(param);
      expect(res, 6);
      expect(param.a, 11);
    });

    test('call: param list', () {
      var lib = new TestLib();
      var res = lib.paramList([1, 2]);
      expect(res, 3);
    });

    test('call: param list out', () {
      var lib = new TestLib();
      var param = new Ref([1, 2]);
      var res = lib.paramListOut(param);
      expect(res, 3);
      expect(param.value[0], 11);
    });

    test('call: param data', () {
      var lib = new TestLib();
      var data = new ByteData(1);
      data.setUint8(0, 1);
      var res = lib.paramTypeData(data);
      expect(res, 1);
    });

    test('call: param data out', () {
      var lib = new TestLib();
      var data = new ByteData(1);
      data.setUint8(0, 1);
      var res = lib.paramTypeDataOut(data);
      expect(res, 1);
      expect(data.getUint8(0), 2);
    });

    test('call: param null', () {
      var lib = new TestLib();
      
      expect(lib.paramNullInt(null), 1);
      expect(lib.paramNullSt(null), 1);
      expect(lib.paramNullString(null), 1);
      expect(lib.paramNullList(null), 1);
    });    

    test('call: struct string', () {
      var lib = new TestLib();
      var param = new NestSt(1, "abc", null, 2);
      var res =  lib.structString(param);
      expect(res, 6);
    });

    test('call: struct nested', () {
      var lib = new TestLib();
      var param = new NestSt(4, "", new IntsSt(1,2,3), 5);
      var res =  lib.structNested(param);
      expect(res, 15);
    });

    test('call: struct nested out', () {
      var lib = new TestLib();
      var param = new NestSt(4, "", new IntsSt(1,2,3), 5);      
      
      var res = lib.structNestedOut(param);
      expect(res, 15);
      expect(param.a, 14);
      expect(param.p.a, 21);
    });

    test('call: struct by value', () {
      var lib = new TestLib();
      var param = new ByValueSt(4, new IntsSt(1,2,3), "", "", new ByteData(4), 6.0, 5);
      var res = lib.structByValue(param);
      expect(res, 9);
    });

    test('call: struct by value null', () {
      var lib = new TestLib();
      var param = new ByValueSt(null, null, null, null, null, null, null);
      var res = lib.structByValue(param);
      expect(res, 0);
    });

    test('call: struct all ints ', () {
      var lib = new TestLib();
      var param = new AllIntsSt(2, 3, 4, 5, 6, 12, 13, 14, 15, 16);
      var res = lib.structAllInts(param);
      expect(res, 90);
    });
    
    test('call: return void', () {
      var lib = new TestLib();      
      lib.returnVoid(1);
    });    

    
    test('typeddata pointer', () {
      var data = new Uint8List(10);
      var pointer1 = getPointer(data);
      var pointer2 = getPointer(data);
      expect(pointer1, pointer2);
    });

    test('native sizes', () {
      var sizes = getSizes();
      expect(sizes, isNotEmpty);
    });

    test('struct encode/decode', () {
      var st = new IntsSt(1, 2, 3);
      var meta = getStructMeta(reflectType(IntsSt));
      var data = encodeStruct(st, null, meta, [], []);
      var list = data.buffer.asInt8List().toList();
      expect(list, [1,0,0,0, 2,0,0,0, 3,0,0,0]);

      var emptySt = new IntsSt(0, 0, 0);
      decodeStruct(emptySt, data, meta, []);
      expect(emptySt.a, equals(1));
      expect(emptySt.b, equals(2));
      expect(emptySt.c, equals(3));
    });

  });
}

@Library('libdna.so', 'dna.dll')
class TestLib extends DynamicLibrary<TestLib> {

  @Method(C_CHAR)
  int paramChar(@Param(C_CHAR) int a, @Param(C_CHAR) int b);
  @Method(C_UCHAR)
  int paramUchar(@Param(C_UCHAR) int a, @Param(C_UCHAR) int b);
  @Method(C_INT)
  int paramInt(@Param(C_INT) int a, @Param(C_INT) int b);
  @Method(C_UINT)
  int paramUint(@Param(C_UINT) int a, @Param(C_UINT) int b);

  @Method(C_FLOAT)
  double paramFloat(@Param(C_FLOAT) double a, @Param(C_FLOAT) double b);
  @Method(C_DOUBLE)
  double paramDouble(@Param(C_DOUBLE) double a, @Param(C_DOUBLE) double b);
  @Method(C_BOOL)
  bool paramBool(@Param(C_BOOL) bool a, @Param(C_BOOL) bool b);

  int paramNullInt(@Param(C_INT) int i);
  int paramNullSt(@Param(C_STRUCT) IntsSt st);
  int paramNullString(@Param(C_STRING) String str);
  int paramNullList(@Param(C_INT) List<int> li);

  int paramString(@Param(C_STRING) String s);
  int paramStringOut(@Param(C_STRING, out: true) Ref<String> str);

  int paramStruct(@Param(C_STRUCT) IntsSt st);
  int paramStructPointer(@Param(C_POINTER) int st);
  int paramStructOut(@Param(C_STRUCT, out: true) IntsSt st);

  int paramList(@Param(C_INT) List<int> li);
  int paramListOut(@Param(C_INT, out: true) Ref<List<int>> li);

  int paramTypeData(@Param(TYPEDDATA) TypedData data);
  int paramTypeDataOut(@Param(TYPEDDATA) TypedData data);
  
  int structString(@Param(C_STRUCT) NestSt st);
  int structNested(@Param(C_STRUCT) NestSt st);
  int structNestedOut(@Param(C_STRUCT, out: true) NestSt st);
  int structByValue(@Param(C_STRUCT) ByValueSt st);
  int structAllInts(@Param(C_STRUCT) AllIntsSt st);
  
  void returnVoid(@Param(C_INT) int a);
}

@Struct()
class IntsSt {
  @Field(C_INT) int a;
  @Field(C_INT) int b;
  @Field(C_INT) int c;

  IntsSt(this.a, this.b, this.c);
}

@Struct()
class NestSt {
  @Field(C_INT) int a;
  @Field(C_STRING) String s;
  @Field(C_STRUCT, ref: true) IntsSt p;
  @Field(C_INT) int b;

  NestSt(this.a, this.s, this.p, this.b);
}

@Struct()
class ByValueSt {
  @Field(C_INT) int a;
  @Field(C_STRUCT) IntsSt p;
  @Field(C_STRING) String s;
  @Field(C_CHAR, length: 10) String ls;
  @Field(C_INT, length: 3) ByteData li;
  @Field(C_DOUBLE) double d;
  @Field(C_INT) int b;

  ByValueSt(this.a, this.p, this.s, this.ls, this.li, this.d, this.b);
}

@Struct()
class AllIntsSt {
  @Field(C_CHAR) int cha;
  @Field(C_SHORT) int sho;
  @Field(C_INT) int intt;
  @Field(C_LONG) int lon;
  @Field(C_LONGLONG) int lonlon;
  @Field(C_UCHAR) int ucha;
  @Field(C_SHORT) int usho;
  @Field(C_UINT) int uintt;
  @Field(C_ULONG) int ulon;
  @Field(C_ULONGLONG) int ulonlon;

  AllIntsSt(this.cha, this.sho, this.intt, this.lon, this.lonlon,
    this.ucha, this.usho, this.uintt, this.ulon, this.ulonlon);
}