//  Copyright 2016 root.ext@gmail.com
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import 'dart:mirrors';
import 'dart:typed_data';

import 'types.dart';
import 'util.dart';
import 'util_c.dart';
import 'log.dart';
import 'model/method_meta.dart';
import 'model/param_meta.dart';
import 'model/field_meta.dart';
import 'model/struct_meta.dart';
import 'model/argument.dart';

List<int> _structTypes(dynamic object, StructMeta structMeta){
  var types = new List<int>();
  types.add(PUSH);

  iterateFields(object, structMeta, (field, fieldMeta, fieldMirror) {
    switch (fieldMeta.type) {
      case C_STRING:
        if (field != null) {
          types.add(PUSH);
          types.add(POP);
          types.add(OFFSET);
        }
        break;
      case C_STRUCT:
        if (fieldMeta.ref) {
          types.addAll(_structTypes(field, fieldMeta.struct));
          types.add(OFFSET);
        }
        break;
    }
  });

  types.add(POP);
  return types;
}

List<int> paramTypes(Invocation invocation, MethodMeta methodMeta) {
  var types = new List<int>();

  var index = 0;
  for (var param in methodMeta.params) {
    var argument = invocation.positionalArguments[index];
    if (argument == null) {
      types.add(TYPEDDATA);
      break;
    }

    if (param.list) {
      types.add(TYPEDDATA);
    } else {
      switch (param.type) {
        case C_STRING:
          types.add(TYPEDDATA);
          break;
        case C_STRUCT:
          types.addAll(_structTypes(argument, param.struct));
          break;
        default:
          types.add(param.type);
      }
    }
    
    index++;
  }

  return types;
}

List<Object> encodeParams(
    Invocation invocation, List<ParamMeta> params, List<Argument> refArgs) {
  var nativeParams = new List<Object>();

  var index = 0;
  for (var param in params) {
    var argument = invocation.positionalArguments[index];
    if (argument == null) {
      nativeParams.add(null);
      break;
    }

    if (param.list) {
      if (param.out) {
        var data = encodeList(argument.value, param.type);
        refArgs.add(new Argument(argument, data));
        nativeParams.add(data);
      } else {
        nativeParams.add(encodeList(argument, param.type));
      }
    } else {
      switch (param.type) {
        case C_STRING:
          {
            if (param.out) {
              var data = stringToData(argument.value);
              refArgs.add(new Argument(argument, data));
              nativeParams.add(data);
            } else {
              nativeParams.add(stringToData(argument));
            }
          }
          break;
        case C_STRUCT:
          {           
            if (param.out) {
              var data = encodeStructRef(argument, param.struct, nativeParams, refArgs);
              refArgs.add(new Argument(argument, data));
            } else {
              encodeStructRef(argument, param.struct, nativeParams, refArgs);
            }
          }
          break;
        default:
          nativeParams.add(argument);
      }
    }
    index++;
  }

  refArgs.forEach((ref) => logt('ref ${ref.data}'));
  return nativeParams;
}

void decodeParams(
  Invocation invocation, List<ParamMeta> params, List<Argument> refArgs) {
  var index = 0;
  for (var param in params) {
    var argument = invocation.positionalArguments[index];
    if (argument == null) {
      break;
    }

    if (param.list) {
      if (param.out) {
        var data = refArgs.where((e) => e.host == argument).first.data;
        argument.value = decodeList(data, param.type);
      }
    } else {
      switch (param.type) {
        case C_STRING:
          {
            if (param.out) {
              var data = refArgs.where((e) => e.host == argument).first.data;
              argument.value = dataToString(data);
            }
          }
          break;
        case C_STRUCT:
          {
            if (param.out) {
              var data = refArgs.where((e) => e.host == argument).first.data;
              argument = decodeStruct(argument, data, param.struct, refArgs);
            }
          }
          break;
      }
    }
    index++;
  }
}

TypedData encodeStructRef(dynamic object, StructMeta structMeta, 
  List<Object> nativeParams, List<Argument> refArgs, { int level = 0}) {
  
  var paramIndex = nativeParams.length;
  nativeParams.add(0);

  var data = encodeStruct(object, null, structMeta, nativeParams, refArgs, level: level);

  nativeParams[paramIndex] = data;
  nativeParams.add(0);

  return data;
}

TypedData encodeStruct(dynamic object, TypedData data, StructMeta structMeta,
  List<Object> nativeParams, List<Argument> refArgs, { int level = 0}) {

  var totalSize = structMeta.size;    
  logd('encode ${object} size ${totalSize}');
  
  data = data ?? new ByteData(totalSize);
  ByteData view = new ByteData.view(data.buffer);
  iterateFields(object, structMeta, (field, fieldMeta, fieldMirror) {
    var type = fieldMeta.type;
    var size = fieldMeta.size;
    var offset = fieldMeta.offset;
    logd('  ${fieldMirror} [${size}, ${offset}]');

    if (fieldMeta.length > 0) {
      if (field == null) {
        return;
      }
      if (field is TypedData) {
        ByteData fieldData = new ByteData.view(field.buffer);
        copyBySrc(fieldData, view, offset);
      } else if (field is String) {
        ByteData fieldData = new ByteData.view(stringToData(field+"\x00").buffer);
        copyBySrc(fieldData, view, offset);
      }
    } else {
      if (isInteger(type)) {
        setInt(view, size, offset, field != null ? (field as int) : 0);
      } else if (isDecimal(type)) {
        setFloat(view, size, offset, field != null ? (field as double) : 0.0);
      }
    }

    switch (type) {
      case C_BOOL:
        setInt(view, size, offset, (field as bool) ? 1 : 0);
        break;
      case C_STRING:
        if (field == null) {
          setInt(view, size, offset, 0);
          break;
        }
        TypedData fieldData = stringToData(field);
        refArgs.add(new Argument(field, fieldData));
        logt('  string field data ${fieldData}');
        nativeParams.add(fieldData);
        nativeParams.add(0);
        nativeParams.add(offset);
        break;
      case C_STRUCT:
        if (fieldMeta.ref) {
          TypedData fieldData = encodeStructRef(field, fieldMeta.struct, 
            nativeParams, refArgs, level: level+1);
          logt('  struct field data ${fieldData}');
          refArgs.add(new Argument(field, fieldData));       
          nativeParams.add(offset);  
        } else {
          encodeStruct(field, data, fieldMeta.struct, 
            nativeParams, refArgs, level: level+1);
        }        
        break;
    }
  });
  logd('encoded ${object}');
  logt('encoded data ${data}');
  return data;
}

dynamic decodeStruct(dynamic object, TypedData data, StructMeta structMeta,
  List<Argument> refArgs) {

  var totalSize = structMeta.size;  
  logd('decode ${object} size ${totalSize}');
  logt('decode data ${data}');

  ByteData view = new ByteData.view(data.buffer);
  InstanceMirror classMirror = reflect(object);

  iterateFields(object, structMeta, (field, fieldMeta, fieldMirror) {
    var type = fieldMeta.type;
    var size = fieldMeta.size;
    var offset = fieldMeta.offset;
    logd('  ${fieldMirror} [${size}, ${offset}]');

    dynamic value;
    if (fieldMeta.length > 0) {
      if (field == null) {
        return;
      }
      if (field is TypedData) {
        ByteData valueData = new ByteData(size);
        copyByDest(view, valueData, offset);
        value = valueData;
      } else if (field is String) {
        ByteData valueData = new ByteData(size);
        copyByDest(view, valueData, offset);
        value = trimCString(dataToString(valueData));
      }        
    } else {
      if (isInteger(type)) {
        value = getInt(view, size, offset);
      } else if (isDecimal(type)) {
        value = getFloat(view, size, offset);
      }
    }

    switch (type) {
      case C_BOOL:
        value = getInt(view, size, offset) == 1;
        break;
      case C_STRING:
        var fieldData = refArgs.where((e) => e.host == field).first.data;
        value = dataToString(fieldData);
        logt('  string field data ${fieldData}');
        break;
      case C_STRUCT:
        if (fieldMeta.ref) {
          var fieldData = refArgs.where((e) => e.host == field).first.data;
          logt(' struct field data ${fieldData}');
          value = decodeStruct(field, fieldData, fieldMeta.struct, refArgs);          
        } else {
          value = decodeStruct(field, view, fieldMeta.struct, refArgs);
        }
        
        break;
    }

    classMirror.setField(fieldMirror.simpleName, value);
  });

  logd('decoded ${object}');
  return object;
}

TypedData encodeList(List object, int type) {
  var size = ctypeToSize(type);
  TypedData data;

  if (isInteger(type)) {
    data = fromIntList(object, size);
  } else if (isDecimal(type)) {
    data = fromFloatList(object, size);
  }
  return data;
}

List decodeList(TypedData data, int type) {
  var size = ctypeToSize(type);
  List list;

  if (isInteger(type)) {
    list = toIntList(data, size);
  } else if (isDecimal(type)) {
    list = toFloatList(data, size);
  }
  return list;
}

void iterateFields(dynamic struct, StructMeta structMeta,
    dynamic process(dynamic field, FieldMeta fieldMeta, DeclarationMirror fieldMirror)) {
  ClassMirror classMirror = reflect(struct).type;

  var fields = classMirror.declarations.values;

  for (var fieldMirror in fields) {
    var fieldMeta = structMeta.fields[MirrorSystem.getName(fieldMirror.simpleName)];
    if (fieldMeta == null) continue;
    
    var field = reflect(struct).getField(fieldMirror.simpleName).reflectee;
    process(field, fieldMeta, fieldMirror);
  }
}