//  Copyright 2016 root.ext@gmail.com
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import 'dart:mirrors';

import 'native.dart';
import 'sizes.dart';
import 'types.dart';
import 'util_reflection.dart';
import 'util_c.dart';
import 'log.dart';
import 'model/library.dart';
import 'model/library_meta.dart';
import 'model/method.dart';
import 'model/method_meta.dart';
import 'model/param.dart';
import 'model/param_meta.dart';
import 'model/field.dart';
import 'model/field_meta.dart';
import 'model/struct_meta.dart';

final int NULL_POINTER = 0;

LibraryMeta getLibraryMeta(Type type, Resolver resolver) {
  LibraryMeta library = new LibraryMeta();
  library.type = type;
  library.resolver = resolver;

  var libTag = findTag(reflectClass(type), Library);
  assert(libTag != null, 'class has no @Library');
  var name = (libTag.getField(#name).reflectee as String);
  var altName = (libTag.getField(#altName).reflectee as String);
  assert(name != null, 'library has no name');

  library.pointer = loadLibrary(name);
  if (library.pointer == NULL_POINTER && altName != null) {
    library.pointer = loadLibrary(altName);
  }
  //if (library.pointer == NULL_POINTER) throw new ArgumentError('cannot load library $name');
  library.methods = new Map<String, MethodMeta>();
  library.structs = new Map<String, StructMeta>();

  return library;
}

MethodMeta getMethodMeta(Invocation invocation, LibraryMeta library) {
  String name = MirrorSystem.getName(invocation.memberName);
  if (library.methods.containsKey(name)) {
    return library.methods[name];
  }

  var classMirror = reflectClass(library.type);
  var methods = findMethods(classMirror, invocation.memberName);
  if (classMirror.superinterfaces.isNotEmpty) {
    methods = findMethods(classMirror.superinterfaces.first, invocation.memberName);
  }

  assert(methods.isNotEmpty, 'cannot find method ${invocation.memberName}');
  var methodMirror = methods.first;

  MethodMeta methodMeta = new MethodMeta();
  methodMeta.name = MirrorSystem.getName(methodMirror.simpleName);

  methodMeta.params = new List<ParamMeta>();
  var paramMirrors = methodMirror.parameters;

  for (var paramMirror in paramMirrors) {
    var paramTag = findTag(paramMirror, Param);
    var param = paramTag != null
        ? (paramTag.reflectee as Param)
        : new Param(dartTypeToCType(paramMirror.type));

    var paramTypeName = MirrorSystem.getName(paramMirror.type.simpleName);    
    var struct = param.type == C_STRUCT
        ? (library.structs.containsKey(paramTypeName) 
            ? library.structs[paramTypeName] 
            : getStructMeta(paramMirror.type, offset: 0))
        : null;

    var isRef = false;
    var isList = false;    
    var paramType = paramMirror.type;
    if (paramType.simpleName == #Ref) {
      isRef = true;
      paramType = paramType.typeArguments.first;
    }
    isList = paramType.simpleName == #List && 
      param.type != LISTSTRING;

    methodMeta.params.add(new ParamMeta(param.type, 
      out: param.out || isRef, list: isList, struct: struct));
  }

  var methodTag = findTag(methodMirror, Method);
  var method = methodTag != null 
    ? (methodTag.reflectee as Method) 
    : new Method(dartTypeToCType(methodMirror.returnType), null);
  
  methodMeta.name = method.name != null ? method.name : name;
  methodMeta.returnParam = new ParamMeta(method.type);
  methodMeta.pointer = resolveMethod(invocation, library, methodMeta.name);

  library.methods[name] = methodMeta;
  return methodMeta;
}

int resolveMethod(Invocation invocation, LibraryMeta library, String name) {
  logd("resolve method ${name}");
  if (library.resolver != null) {
    var pointer = library.resolver(invocation);
    if (pointer == NULL_POINTER) {
      pointer = findSymbol(library.pointer, name);
    }
    return pointer;
  }

  var methodPointer = findSymbol(library.pointer, name);
  return methodPointer;
}

StructMeta getStructMeta(TypeMirror typeMirror, {int offset = 0}) {
  StructMeta struct = new StructMeta();
  Map<String, FieldMeta> fields = new Map();

  var sizes = new Sizes();    
  for (var fieldMirror in findDeclarations(typeMirror)) {
    FieldMeta fieldTag = getFieldMeta(fieldMirror);
    if (fieldTag == null) continue;

    var fieldSize = fieldTag.size;
    sizes.alignField(fieldSize);
    
    var fieldOffset = sizes.offset;    
    sizes.addField(fieldSize);        
    
    StructMeta fieldStructMeta = null;  
    if (fieldTag.type == C_STRUCT) {
      fieldStructMeta = getStructMeta((fieldMirror as VariableMirror).type, 
        offset: fieldTag.ref ? 0 : sizes.offset);      
      if (!fieldTag.ref) {
        sizes.alignField(fieldStructMeta.maxSize);
        sizes.addStruct(fieldStructMeta.size, fieldStructMeta.maxSize);
      }
    }

    FieldMeta fieldMeta = new FieldMeta(fieldTag.type, 
      fieldTag.ref, fieldTag.length,
      fieldSize, fieldOffset + offset);
    fieldMeta.struct = fieldStructMeta;

    String name = MirrorSystem.getName(fieldMirror.simpleName);  
    fields[name] = fieldMeta;    
  }
  sizes.alignByMax();
  
  struct.fields = fields;
  struct.size = sizes.totalSize;  
  struct.maxSize = sizes.maxSize;
  return struct;
}

FieldMeta getFieldMeta(DeclarationMirror fieldMirror) {
  var fieldDecl = findTag(fieldMirror, Field);
  if (fieldDecl == null) return null;

  Field fieldTag = fieldDecl.reflectee;
  var type = fieldTag.type;
  var ref = fieldTag.ref;
  var size = ctypeToSize(type, ref: ref);
  if (fieldTag.length > 0) {
    size = size * fieldTag.length;
  }

  return new FieldMeta(type, ref, fieldTag.length, size, 0);
}
